##############################################################################
# deldenoiser is a python package supporting the deldenoiser command line tool
# which is used to denoise read counts of a DNA-encoded experiment.
#
# Copyright (C) 2020  Totient, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v3.0
# along with this program.
# If not, see <https://www.gnu.org/licenses/gpl-3.0.en.html>.
#
# Developer:
# Peter Komar (peter.komar@totient.bio)
##############################################################################

import pytest

import os

try:
    from deldenoiser import nullblockmodel as nbm
    from deldenoiser import pbs_algorithm as pbs
except ImportError:
    import sys

    sys.path.append(
        os.path.dirname(
            os.path.dirname(os.path.realpath(__file__))))
    from deldenoiser import nullblockmodel as nbm
    from deldenoiser import pbs_algorithm as pbs

import numpy as np


def is_approx(A, B, rel_error=1e-8, abs_error_floor=1e-10):
    A = np.array(A)
    B = np.array(B)
    if A.shape == B.shape and A.size == 0:
        return True
    if not (np.isinf(A) == np.isinf(B)).all():
        return False
    if not (A[np.isinf(A)] == B[np.isinf(B)]).all():
        return False
    A = A[~np.isinf(A)]
    B = B[~np.isinf(B)]
    diff = np.abs(A - B)
    size = np.max(np.abs([A, B]), axis=0)
    return (diff <= rel_error * size + abs_error_floor).all()


@pytest.mark.parametrize(
    "s, x, expected_result",
    [
        (np.array([1, 1, 2]), np.array([0.0, 0.5, 0.5]),
         np.array([0.0,
                   -0.5,
                   -0.09453489189183561802198688453565086342800957653750580238]
                  )
         )
    ]
)
def test_gammainccln(s, x, expected_result):
    assert is_approx(pbs.gammainccln(s, x), expected_result)


def test_NullblockModel():
    design = [10, 10, 10]
    alpha = 1.0
    gamma = 1.0
    yields = [np.array([0.8] * 10)] * 3
    path, file = os.path.split(os.path.realpath(__file__))
    Npost_file = os.path.join(path, 'Npost.tsv')

    model = nbm.NullBlockModel(design, alpha, gamma)

    model.yields = yields
    model.x = model._compute_x()
    model.load_postselection_readcount(Npost_file)

    model.guess_truncates(processes=2)
    model.fit_truncates()
    F = model.fit_fullcycles()
    model.compute_clean_readcounts(F)



##############################################################################
# Implementation of the Null-block model
#
#
# Stored attributes:
#
#   cycles: number of synthesis cycles
#   design: list of lane numbers in each cycle, r_c^{max}
#   alpha: L1 regularization parameter
#   gamma: dispersion parameter
#   bhat: sequence imbalance estimates, bhat_{c, r_c}
#   yields: yields of synthesis reactions, Y_{c, r_c}
#   x: x_{c, q_c} = bhat_{c,q_c} Y_{c, q_c},  if q_c > 0
#      x_{c, 0} = 1.0 - sum_{r_c} bhat_{c, r_c} Y_{c, r_c}
#   r_arr: array of reaction index vectors, 1 <= r_c <= r_c^{max}
#       for which N_r > 0
#   n: vector of scaled post-selection readcounts, n_r = N_r / gamma
#   ntot: total scaled readcount
#   Z: vector of Z_{r} = alpha + X_{r,r}/gamma,
#      where X_{r,r} = N_tot * prod_{c} x_{c, r_c}
#   a: background intensity, a_r = sum_q b_r[q] * F_q
#      where b_r[q] = (alpha + X_{r,r} / gamma)
#                     * prod_{c: q_c==0} (1 - yields[c,r_c]) / yields[c,r_c]
#   q_arr: array of truncates indexes which have potentially F_q > 0
#   F: vector of fitness coefficients of the truncates, F_q
#   c0: c0[q] = - alpha + sum_{r} b_r[q]
#               - ntot * prod_{c} x_{c, q_c}
#       where b_r[q] = Z_{r}
#                      * prod_{c: q_c==0} (1 - yields[c,r_c]) / yields[c,r_c]
#
#
# Functions implementing the following steps:
#
#   1. fit_sequence_imbalance(): Read pre-selection read count data,
#       line-by-line, and compute bhat
#   2. load_yields(): Load yields
#   3. load_postselection_readcount(): Load post-selection read count data
#       entirely, and compute Z and a
#   4. guess_truncates(): Find truncates (q) that are non-zero
#       and compute initial guesses
#   5. fit_truncates(): Fit fitness of truncates
#   6. fit_fullcycles(): Fit full-cycle product fitness coefficients
#   7. compute_clean_readcounts(): Compute clean readcounts
#       for full-cycle products
#
#
# Multiprocessing considerations:
#
#   The most time-consuming steps are step 4 and 5.
#   They can be sped up by distributing the tasks for different q truncates
#   to different processes.
#   The processes share the following arrays: r_arr, n, Z, a
#   To save memory,
#   these arrays must be initialized using the following recipe:
#   >>> import multiprocessing as mp
#   >>> import ctypes
#   >>> import numpy as np
#   >>> base = mp.Array(ctypes.c_double, <size>)  # or with `ctypes.c_int`
#   >>> shared_array = np.ctypeslib.as_array(base.get_obj())
#   >>> shared_array = shared_array.reshape(<shape>)
#   Furthermore, mp.Array() returns a multiprocessing.RLock object,
#   which is used to block overlapping edits on array a
#
#
# Copyright (C) 2020  Totient, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v3.0
# along with this program.
# If not, see <https://www.gnu.org/licenses/gpl-3.0.en.html>.
#
# Developer:
# Peter Komar (peter.komar@totient.bio)
##############################################################################

import sys
import gzip
import numpy as np
import pandas as pd
import multiprocessing as mp
from itertools import product, chain, cycle
import ctypes
from warnings import warn
from deldenoiser.pbs_algorithm import pbs_optimize, gammainccln

##############################################################################
# Globals ####################################################################
##############################################################################
# These arrays must be stored globally, because they are accessed by functions
# below that also must be defined globally for the multiprocessing starmap()
# to be able to distribute them on pool workers.
#
# Because of this design choice at most one instance of NullBlockModel class
# should be created. Constructing NullBlockModel object resets these globals.

global_r_arr = None
global_n = None
global_Z = None
global_a = None
global_a_lock = None

NUMERICAL_FLOOR = 1e-10

##############################################################################


def _read_gzip_or_nogzip(path):
    if path.endswith('.gz'):
        with gzip.open(path, mode='rt', encoding='utf-8') as f:
            for line in f:
                yield line
    else:
        with open(path, 'r') as f:
            for line in f:
                yield line


def _create_shared_array(shape, ctype, lock=True):
    """Create a numpy array that can be shared between parallel processes.

    Args:
        shape (iterable of int): shape of numpy array
        ctype (ctypes type): dtype of numpy array
        lock (bool): if True, the array has multiprocessing lock

    Returns:
        numpy.ndarray: numpy array referring to the data in shared memory
        multiprocessing.synchronize.RLock or None: lock object associated with
            the array
    """
    base = mp.Array(ctype, int(np.prod(shape)), lock=lock)
    if lock:
        arr_lock = base.get_lock()
        base = base.get_obj()
    else:
        arr_lock = None
    arr = np.ctypeslib.as_array(base)
    arr = arr.reshape(shape)
    return arr, arr_lock


def _guess_one_truncate(q, cycles, alpha, ntot, x, yields):
    c0, b, n_median = NullBlockModel._compute_c0_b_nmedian(
        q, cycles, alpha, ntot, x, yields)
    if c0 < 0.0:
        return q, 0.0, c0
    guess = len(b) * n_median / float(np.sum(b) - c0)
    return q, guess, c0


def _fit_one_truncate(q_idx, q, Fq, c0, yields, inner_itermax, tol):
    n, a, b, idxs = NullBlockModel._compute_n_a_b_idx(q, Fq, yields)
    new_Fq = pbs_optimize(n, a, b, c0,
                          beta_start=Fq,
                          itermax=inner_itermax,
                          b_beta_tol=tol)
    global_a_lock.acquire()
    if Fq is None:
        global_a[idxs] += b * new_Fq
    else:
        global_a[idxs] = np.max(np.stack([np.ones_like(b) * NUMERICAL_FLOOR,
                                          global_a[idxs] + b * (new_Fq - Fq)
                                          ]),
                                axis=0)
    global_a_lock.release()
    return q_idx, new_Fq


class NullBlockModel:
    """Implements the Null-block model for denoising DEL readcounts

    Attributes:
        cycles (int): number of synthsis cycles
        design (numpy.ndarray): 1-d array of positive integers,
            each indicating the number of synthesis lanes in the cycles
        alpha (float): positive regularization constant,
            equivalent to L1 (lasso) regularization constant
        gamma (float): positive dispersion constant,
            gamma == 1 means a Poisson noise is assumed
            gamma > 1 means an over-dispersed Poisson distribution is assumed
            gamma < 1 means an under-dispersed Poisson distribution is used
        bhat (list of numpy.ndarray): sequence imbalance factors,
            each bhat[c][r_c] stand for cycle c, reaction r_c
        yields (list of numpy.ndarray): reaction yields,
            each yields[c][r_c] stands for cycle c, reaction r_c
        x (list of numpy.ndarray): intermediary variable
            x = {x[c][q_c]: c = 1, 2, len(design), q_c = 0, 1, ... design[c]}
            x[c][q_c] = bhat[c][q_c] * yields[c][q_c], if q_c != 0
            x[c][q_c] = sum_{r_c} bhat[c][r_c]*(1 - yields[c][r_c]), if q_c==0
        r_arr (numpy.ndarray): 2-d integer array containing
            the reaction indexes
            of the non-zero counts in the post-selection read count data
        n (numpy.ndarray): 1-d integer array containing the non-zero
            scaled counts, n = N / gamma
        ntot (int): total scaled counts
        Z (numpy.ndarray): 1-d array, Z_{r} = alpha + X_{r,r}/gamma,
            where X_{r,r} = N_tot * prod_{c} x[c][r_c]
        a (numpy.ndarray): 1-d array, a_r = sum_q b_r[q] * F_q
            where
            where b_r[q] = Z_{r} *
                prod_{c: q_c==0} (1 - yields[c,r_c]) / yields[c,r_c]
            and F_q is the fitness coefficient of truncate q
        q_arr (numpy.ndarray): 2-d array of truncate indexes
            where only those are stored that have potentially non-zero F_q
        F (numpy.ndarray): 1-d array, F_q fitness coefficients,
            for only those q, that are stored in q_arr
        c0 (numpy.ndarray): 1-d array, sum_{r} X_{r,q}, intermediary array
            c0[q] = - alpha + sum_{r} b_r[q]
                    - ntot * prod_{c} prod_{c} x[c][q_c]
    """
    def __init__(self, design, alpha=1.0, gamma=1.0, default_yield=0.5):
        # check input arguments
        NullBlockModel._check_design(design)
        NullBlockModel._check_positive_float(alpha, 'alpha')
        NullBlockModel._check_positive_float(gamma, 'gamma')

        # load attributes directly from input arguments
        self.cycles = len(design)
        self.design = design
        self.alpha = float(alpha)
        self.gamma = float(gamma)

        # load default values for bhat, yields and x
        self.bhat = [np.ones(lanes, dtype=float) / lanes for lanes in design]
        self.yields = [default_yield * np.ones(lanes, dtype=float)
                       for lanes in design]
        self.x = self._compute_x()

        # tie r_arr, n, Z, a to the globals and reset them
        global global_r_arr
        global_r_arr = None
        self.r_arr = global_r_arr

        global global_n
        global_n = None
        self.n = global_n

        global global_Z
        global_Z = None
        self.Z = global_Z

        global global_a
        global_a = None
        self.a = global_a

        global global_a_lock
        global_a_lock = None

        # initialize ntot, q_arr, F, c0 to empty values
        self.ntot = 0.0
        self.q_arr = np.zeros([0, self.cycles], dtype=int)
        self.F = np.zeros(0, dtype=float)
        self.c0 = np.zeros(0, dtype=float)

    @staticmethod
    def _check_design(design):
        error_message = 'design must be a sequence of positive integers, ' \
            f'value passed: {design}'
        try:
            d_arr = np.array(design)
            if len(d_arr.shape) != 1:
                raise ValueError(error_message)
            if len(d_arr) == 0:
                raise ValueError(error_message)
            if np.sum(d_arr <= 0) > 0:
                raise ValueError(error_message)
        except ValueError:
            raise ValueError(error_message)

    @staticmethod
    def _check_positive_float(arg, name):
        error_message = f'{name} must be a positive floating point number, ' \
            f'value passed: {arg}'
        try:
            f = float(arg)
            if f <= 0.0:
                raise ValueError(error_message)
        except ValueError:
            raise ValueError(error_message)

    def _compute_x(self):
        cycles = self.cycles
        bhat = self.bhat
        yields = self.yields
        x = [np.concatenate((np.array([0]),
                             bhat[c] * yields[c])) for c in range(cycles)]
        for c in range(cycles):
            x[c][0] = 1.0 - np.sum(x[c])
        return x

    def _check_read_count_file(self, read_count_file):
        expected_header = [f'cycle_{c+1}_lane' for c in range(self.cycles)] \
            + ['readcount']
        line = _read_gzip_or_nogzip(read_count_file).__next__()
        header = line.strip().split('\t')
        if header != expected_header:
            error_message = f'The header {expected_header} is expected, '\
                f'but not found in {read_count_file}. ' \
                f'Header found: {header}'
            raise ValueError(error_message)

    def fit_sequence_imbalance(self, pre_selection_read_count_file):
        """Fit the sequence imbalance factors, bhat[c][r_c] from data file

        Using the formula:
            ```
            bhat[c][r_c] = (1 + sum_{r': r'_c == r_c} N[r']) /
                           (N0[c] +  sum_{r'} N[r'])
            ```
        where N[r'] is the number of pre-selection read count for index r',
        and N0[c] is the number of different r_c values available for cycle c.
        This assumes flat priors for all {bhat[c][r_c]: r_c = 1..r_c^{max}}

        The input file is not loaded to memory, instead it is streamed.
        This allows computing the bhat factors from very big DEL libraries.

        Does not return anything. It updates self.bhat and self.x.

        Args:
            pre_selection_read_count_file (str): file path

        """
        self._check_read_count_file(
            pre_selection_read_count_file)

        cycles = self.cycles
        design = self.design
        bhat = [np.ones(lanes, dtype=float) for lanes in design]
        Ntot = 0
        fin = _read_gzip_or_nogzip(pre_selection_read_count_file)
        fin.__next__()  # skip header

        for line_num, line in enumerate(fin):
            try:
                record = list(map(int, line.strip().split('\t')))
            except ValueError:
                warn(f'Line {line_num + 1} cannot be processed, '
                     f'it is skipped: {line}')
                continue
            if len(record) != cycles + 1:
                warn(f'Line {line_num + 1} contains wrong number of entries, '
                     f'it is skipped: {line}')
                continue

            r = record[:-1]
            Nr = record[-1]
            if Nr > 0:
                for c in range(cycles):
                    bhat[c][r[c] - 1] += Nr
                Ntot += Nr

        for c in range(cycles):
            bhat[c] /= float(design[c] + Ntot)

        self.bhat = bhat
        self.x = self._compute_x()

    def _check_yields_file(self, yields_file, minyield, maxyield):
        if yields_file.endswith('.gz'):
            compression = 'gzip'
        else:
            compression = None
        df_yields = pd.read_csv(yields_file, sep='\t',
                                compression=compression)
        if list(df_yields.columns) != ['cycle', 'lane', 'yield']:
            raise ValueError('yields file must have columns '
                             '"cycle", "lane", "yields')

        error_message = '"cycle" column of yields file ' \
                        'must contain positive integers'
        if ((df_yields['cycle'].values !=
             df_yields['cycle'].values.astype(int)).any()):
            raise ValueError(error_message)
        if ((df_yields['cycle'].values <= 0).any()):
            raise ValueError(error_message)

        error_message = '"lane" column of yields file ' \
                        'must contain positive integers'
        if ((df_yields['lane'].values !=
             df_yields['lane'].values.astype(int)).any()):
            raise ValueError(error_message)
        if (df_yields['lane'].values <= 0).any():
            raise ValueError(error_message)

        error_message = '"yield" column of yields file must contain ' \
                        'floating point values, 0.0 < yield < 1.0'
        df_yields.loc[df_yields['yield'] <= minyield, 'yield'] = minyield
        df_yields.loc[df_yields['yield'] >= maxyield, 'yield'] = maxyield
        if (df_yields['yield'] <= 0.0).any() or (
                df_yields['yield'] >= 1.0).any():
            raise ValueError(error_message)
        if np.isnan(df_yields['yield']).any():
            raise ValueError(error_message)

        df_yields.sort_values(by=['cycle', 'lane'], inplace=True)
        cycles = df_yields['cycle'].values
        lanes = df_yields['lane'].values
        yields = df_yields['yield'].values

        error_message = '"cycle" and "lane" columns in yields file ' \
                        'must list every cycle-lane ' \
                        'combination in the design in a sorted order.'
        cmax = self.cycles
        design = self.design
        expected_cycles = np.concatenate([(c + 1) * np.ones(design[c])
                                          for c in range(cmax)])
        expected_lanes = np.concatenate([np.arange(1, design[c] + 1, 1)
                                         for c in range(cmax)])
        if (cycles != expected_cycles).any() or (
                lanes != expected_lanes).any():
            raise ValueError(error_message)

        return yields

    def load_yields(self, yields_file, minyield, maxyield):
        """Check and load the reaction yields from file.

        Nothing is returned, self.yields and self.x gets updated.

        Args:
            yields_file (str): file path
            minyield (float): lowest allowed yield value, lower inputs are
                censored to this level
            maxyield (float): highest allowed yield value, higher inputs are
                censored to this level

        """
        cycles = self.cycles
        design = self.design
        yields = self._check_yields_file(yields_file, minyield, maxyield)
        yield_list = []
        idx_start = 0
        for c in range(cycles):
            yield_list.append(yields[idx_start:idx_start + design[c]])
            idx_start += design[c]

        self.yields = yield_list
        self.x = self._compute_x()

    def load_postselection_readcount(self,
                                     post_selection_read_count_file,
                                     index_dtype=ctypes.c_int16,
                                     count_dtype=ctypes.c_int32,
                                     float_dtype=ctypes.c_float):
        """Check and load post-selection read count data.

        Only records that contain positive read counts are kept in memory.
        Data is loaded into multiprocessing Arrays, so they
        can be concurrently accessed by parallel processes.

        Nothing is returned, self.r_arr, self.n, self.Z, self.a get updated.

        Args:
            post_selection_read_count_file (str): file path
            index_dtype (ctypes type): C type of r index
            count_dtype (ctypes type): C type of N readcounts
            float_dtype (ctypes type): C type of n = N/gamma

        """

        self._check_read_count_file(post_selection_read_count_file)

        cycles = self.cycles
        index_cols = [f'cycle_{c+1}_lane' for c in range(cycles)]
        dtypes = dict(zip(index_cols, [index_dtype] * cycles))
        dtypes['readcount'] = count_dtype
        if post_selection_read_count_file.endswith('gz'):
            compression = 'gzip'
        else:
            compression = None
        df_N = pd.read_csv(post_selection_read_count_file, sep='\t',
                           low_memory=True, dtype=dtypes,
                           error_bad_lines=False, warn_bad_lines=True,
                           compression=compression)
        if np.sum(df_N['readcount'] == 0) > 0:
            df_N = df_N[df_N['readcount'] > 0]

        records = len(df_N)

        global global_r_arr
        global_r_arr, _ = _create_shared_array([records, cycles],
                                               index_dtype, lock=False)
        self.r_arr = global_r_arr
        self.r_arr[:, :] = df_N[index_cols].values

        global global_n
        global_n, _ = _create_shared_array([records], float_dtype, lock=False)
        self.n = global_n
        self.n[:] = df_N['readcount'].values / float(self.gamma)
        self.ntot = np.sum(self.n)

        global global_Z
        global_Z, _ = _create_shared_array([records], float_dtype, lock=False)
        self.Z = global_Z
        self.Z[:] = self.ntot
        for c in range(cycles):
            self.Z *= self.x[c][self.r_arr[:, c]]
        self.Z += self.alpha

        global global_a
        global global_a_lock
        global_a, global_a_lock = _create_shared_array([records],
                                                       float_dtype, lock=True)
        self.a = global_a
        self.a += NUMERICAL_FLOOR

    @staticmethod
    def _compute_c0_b_nmedian(q, cycles, alpha, ntot, x, yields):
        """Computes the c0 and b parameters and the median n.

        c0 = -alpha + sum_r(b) - ntot * prod_{c} x[c][q_c]
        where b_r = Z_r * prod_{c: q_c==0} (1 - yields[c,r_c]) / yields[c,r_c]

        Args:
            q (iterable): truncate index vector
            guess (bool): if True, return a guess for

        Returns:
            (float, numpy.ndarray, float): c0, b, median(n)

        """
        q = np.array(q)
        qc_is_zero = (q == 0)
        truncated_c_idxs = np.argwhere(qc_is_zero)[:, 0]
        non_truncated_c_idxs = np.argwhere(~qc_is_zero)[:, 0]

        r_arr_aff = global_r_arr
        Z_aff = global_Z
        n_aff = global_n
        for c in non_truncated_c_idxs:
            condition = (r_arr_aff[:, c] == q[c])
            r_arr_aff = r_arr_aff[condition, :]
            Z_aff = Z_aff[condition]
            n_aff = n_aff[condition]
        b = Z_aff.copy()
        for c in truncated_c_idxs:
            b *= (1.0 / yields[c][r_arr_aff[:, c] - 1] - 1)
        c0 = (- alpha
              + np.sum(b)
              - ntot * np.prod([x[c][q[c]] for c in range(cycles)])
              )
        if len(n_aff) == 0:
            n_median = 0
        else:
            n_median = np.median(n_aff)

        return c0, b, n_median

    def guess_truncates(self, processes=mp.cpu_count()):
        """Determine which q truncates have non-zero F_q, and provide guesses

        The sign of the c0 parameter is used to determine if a truncate has
        non-zero optimal fitness.

        For those that are selected, the following guess is computed
            F_q^{guess} = N_aff * median(N_aff) / float(np.sum(b) - c0)
            where
                c0 = -alpha + sum_r(b) - (1/gamma) * Ntot * prod_c(x[c][q_c])
                b_r = (alpha + X_{r,r} / gamma)
                      * prod_{c: q_c==0} (1 - yields[c,r_c]) / yields[c,r_c]

        Nothing is returned. The arrays self.q_arr, self.F and self.c0
        get updated with values corresponding to q truncates that are deemed
        to have non-zero F_q fitness.

        Args:
            processes (int): number of parallel processes to start

        """
        cycles = self.cycles
        design = self.design
        alpha = self.alpha
        ntot = self.ntot
        x = self.x
        yields = self.yields

        # generate qs, the iterator for all q truncate indexes
        q_iters = []
        for mask in product(*[(0, 1)] * cycles):
            if sum(mask) == cycles:
                continue
            qc_values = []
            for m, d in zip(mask, design):
                if m == 0:
                    qc_values.append((0,))
                else:
                    qc_values.append(tuple(range(1, d + 1)))
            q_iter = product(*qc_values)
            q_iters.append(q_iter)
        qs = chain(*q_iters)

        pool = mp.Pool(processes=processes)
        # syntax: _guess_one_truncate(q, cycles, alpha, ntot, x, yields)
        results = pool.starmap(_guess_one_truncate,
                               zip(qs, cycle([cycles]), cycle([alpha]),
                                   cycle([ntot]), cycle([x]),
                                   cycle([yields])))
        pool.close()

        results_nonzero = []
        for res in results:
            q, Fq, c0 = res
            if Fq == 0.0:
                continue
            results_nonzero.append(res)
        results_nonzero.sort(key=lambda t: t[0])
        self.q_arr = np.array([res[0] for res in results_nonzero])
        self.F = np.array([res[1] for res in results_nonzero])
        self.c0 = np.array([res[2] for res in results_nonzero])

    @staticmethod
    def _compute_n_a_b_idx(q, Fq, yields):
        if Fq is None:
            Fq = 0.0
        q = np.array(q)
        qc_is_zero = (q == 0)
        truncated_c_idxs = np.argwhere(qc_is_zero)[:, 0]
        non_truncated_c_idxs = np.argwhere(~qc_is_zero)[:, 0]

        r_arr_aff = global_r_arr
        idx_aff = np.arange(r_arr_aff.shape[0])
        for c in non_truncated_c_idxs:
            condition = (r_arr_aff[:, c] == q[c])
            r_arr_aff = r_arr_aff[condition, :]
            idx_aff = idx_aff[condition]

        Z_aff = global_Z[idx_aff]
        n_aff = global_n[idx_aff]
        a_aff = global_a[idx_aff].copy()

        # Compute b_r = (alpha_s + X_{r,r}/gamma)
        #               * prod_{c: q_c = 0} (1-Y_{c,q_c}) / Y_{c, q_c}
        b = Z_aff.copy()
        for c in truncated_c_idxs:
            b *= (1.0 / yields[c][r_arr_aff[:, c] - 1] - 1)

        # Compute a_r = (alpha_s / X_{r,r,}  +  1 / gamma) * B_r - b_r F_q
        # but also make sure it's never negative
        a = np.max(np.stack((np.ones_like(a_aff) * NUMERICAL_FLOOR,
                             a_aff - b * Fq)), axis=0)

        return n_aff, a, b, idx_aff

    def fit_truncates(self, outer_itermax=100, inner_itermax=10, tol=0.5,
                      processes=mp.cpu_count(), debug=False, max_downsteps=5,
                      F_init=None):
        """Optimize the fitness of truncates iteratively.

        Args:
            outer_itermax (int): max number of iteration loops over truncates
            inner_itermax (int): max number of N-R or A-B iteration
            tol (float): iterations terminate if self.a changes less than this
                between iterations
            processes (int): number of parallel processes to start
            debug (bool): if True, print messages to stderr
            max_downsteps (int): if log-likelihood decreases
                this number of times throughout the optimization,
                the optimization is terminated
            F_init (None or float): staring guess for fitness values,
                if None the default guess of pbs_optimize() is used

        Nothing is returned. Arrays self.F and self.a are updated.

        """

        # setting initial F values to None makes pbs_algorithm compute the
        # same initial staring guess as self.guess_truncates()
        #
        if F_init is not None:
            F_init = float(F_init)
        Fqs = np.array([F_init for _ in range(self.q_arr.shape[0])])

        pool = mp.Pool(processes=processes)
        q_idxs = np.arange(self.q_arr.shape[0])

        a_curr = self.a.copy()
        logL_prev = -np.inf
        downward_steps = 0
        break_next = False
        for it in range(outer_itermax):
            results = pool.starmap(
                _fit_one_truncate,
                zip(q_idxs, self.q_arr, Fqs, self.c0, cycle([self.yields]),
                    cycle([inner_itermax]), cycle([tol / 2**self.cycles])))
            for res in results:
                q_idx, Fq = res
                Fqs[q_idx] = Fq
            max_change = np.max(np.abs(a_curr - self.a))
            a_curr[:] = self.a
            self.F = Fqs
            logL = self.log_likelihood()
            if debug:
                print(f'iteration {it+1}, logL: {logL}',
                      file=sys.stderr)
            if (max_change < tol) or break_next:
                break
            if logL < logL_prev:
                downward_steps += 1
            logL_prev = logL
            if downward_steps > max_downsteps:
                break_next = True
        pool.close()

    def fit_fullcycles(self):
        """Compute the maximum likelihood estimate of full-cycle fitnesses.

        hat F_r = max(0, N_r / (X_{r,r} + gamma * alpha) - hat B_r / X_{rr} )
        = max(0, (n_r - a_r) / Z_r )

        Returns:
            numpy.ndarray: estimated F_r fitness values for all r reactions
            that have non-zero counts

        """

        F = np.max(np.stack([
            np.zeros_like(self.n),
            (self.n - self.a) / self.Z
        ]), axis=0)
        return F

    def compute_clean_readcounts(self, F):
        """Compute the expectation value of N_{r,r}, given N_r and F.

        N_{r,r} = N_r * (X_{r,r} F_r) / sum_{q} X_{r,q} F_q
        = gamma * n_r * (Z_r F_r) / (Z_r F_r + a_r)

        Args:
            F (numpy.ndarray): fitness of full cycle products, computed by
                self.fit_fullcycles()

        Returns:
            numpy.ndarray: expectation value of N_{r,r}, i.e. the number of
            reads that are correctly attributed to the full cycle product r.

        """
        N_clean_ev = self.gamma * self.n * self.Z * F / (self.Z * F + self.a)
        return N_clean_ev

    def log_likelihood(self):
        logL = 0.0
        logL += np.sum(self.c0 * self.F)
        logL += np.sum(gammainccln(self.n + 1, self.a))
        return logL



__author__ = 'Peter Komar "\
             "<peter.komar@totient.bio>'
__copyright__ = '2020 Totient, Inc'
__version__ = '2.0.0'

import io
from datetime import datetime
from setuptools import setup, find_packages

setup(
    name='deldenoiser',
    version=__version__,
    description='Denoise sequencing data from DEL screens.',
    long_description=io.open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: POSIX',
    ],
    author='Peter Komar',
    author_email='peter.komar@totient.bio',
    maintainer='Peter Komar',
    maintainer_email='peter.komar@totient.bio',
    url='https://github.com/totient-bio/deldenoiser',
    license='Copyright (c) {} Totient, Inc.'.format(
        datetime.now().year
    ),
    packages=find_packages(),
    install_requires=io.open('requirements.txt').read().splitlines(),
    include_package_data=True,
    scripts=["command-line-tool/deldenoiser"],
    python_requires='>=3.6'
)



##############################################################################
# deldenoiser is a python package supporting the deldenoiser command line tool
# which is used to denoise read counts of a DNA-encoded experiment.
#
# Copyright (C) 2020  Totient, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v3.0
# along with this program.
# If not, see <https://www.gnu.org/licenses/gpl-3.0.en.html>.
#
# Developer:
# Peter Komar (peter.komar@totient.bio)
##############################################################################



##############################################################################
# Implementation of the Poisson Background Subtraction algorithm
# It minimizes the log-likelihood of the background intensity beta
#
# logL(beta) = c0 * beta + sum_{r: n_r > 0} logQ (n_r + 1,  a_r + b_r beta)
#
# where
#   c0 can be positive or negative,
#   n_r is positive,
#   a_r and b_r are non-negative
#   logQ() is np.log(scipy.special.gammaincc())
#
# Optimization is done on [0, inf) interval.
# If c0 <= 0, the optimal beta is 0.0, this is quick to return.
# If c0 > 0, the optimum exists, it is unique and
# is guaranteed to be between 0.0 and beta_1
# where beta_1 = (sum_{r: n_r > 0} n_r) / ((sum_{r: n_r > 0} b_r) - c0)
#
# If no starting beta (beta_start) is provided, the guess beta_0 is used
# beta_0 = (#[r: n_r > 0] * median_{r: n_r > 0}(n_r))
#          / (sum_{r: n_r > 0} b_r - c0)
#
# Steps of the algorithm:
# 1. Run Newton-Raphson algorithm, using the 1st and second
#    derivative of logL
#    f(beta) = d/(d beta) logL =
#              c0 - sum_{r: n_1 > 0} b_r H_r
#    f'(beta) = d/(d beta) f =
#               - sum_{r: n_1 > 0} (b_r)^2 H_r ((n_r / xi_r) + H_r - 1)
#    i.e. beta_new = beta_old - f(beta_old) / f'(beta_old)
# where
#    H_r = (xi_r)^(n_r) * exp(-xi_r) / Gamma(n_r + 1, xi_r)
#    xi_r = a_r + b_r beta
#    Gamma(s, x) is th upper incomplete gamma function,
#    i.e. scipy.special.gamma(s) * scipy.special.gammaincc(s,x)
#
# 2. If the Newton-Raphson update fails to decrease |f|, then fall back to
#    Anderson-Bjoerk algorithm on an interval determined by
#    the following rules:
#    - evaluate f at 0, beta_1 and beta_start
#    - if the Newton update (beta') falls between 0 and beta_1,
#      evaluate f(beta')
#    - use the 3 (or 4) points to find the sub-interval
#      on which f is guaranteed tp change sign
#
# 3. Run Anderson-Bjoerck algorithm,
#
#
# Copyright (C) 2020  Totient, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v3.0
# along with this program.
# If not, see <https://www.gnu.org/licenses/gpl-3.0.en.html>.
#
# Developer:
# Peter Komar (peter.komar@totient.bio)
##############################################################################

import numpy as np
from scipy.special import gammaln, gammaincc

NUMERICAL_FLOOR = 1e-10


def gammainccln(s, x):
    """Computes numpy.log(scipy.special.gammaincc(s,x)), avoiding underflow.

    See: https://en.wikipedia.org/wiki/Incomplete_gamma_function#Asymptotic_behavior  # noqa

    Args:
        s (numpy.ndarray): first argument of incomplete gamma function
        x (numpy.ndarray): second argument of incomplete gamma function

    Returns:
        numpy.ndarray: ln(gammaincc(s,x)) for arrays s, x of the same shape
    """
    result = gammaincc(s, x)

    # For some inputs scipy.special.gammaincc returns 0.0.
    # This happens when x >> s.
    # Let's avoid taking the log of these entries,
    # and evaluate the first 3 terms of the asymptotic series instead.
    underflow = (result == 0.0)
    result[~underflow] = np.log(result[~underflow])
    x = x[underflow]
    s = s[underflow]
    result[underflow] = -x + (s - 1) * np.log(x) + np.log(1 + (s - 1) / x) \
                        - gammaln(s)
    return result


def anderson_bjoerk_step(f, start, end, last_yc):
    """Performs a single iteration of Anderson-Bjoerck root-finding algorithm

    Args:
        f: univariate function the root of which we are looking for
        start: x and y values on the left of the interval, tuple of (a, ya)
        end: x and y values on the right of the interval, tuple of (b, yb)
        last_yc: value of the previously discovered y value, yc

    Returns:
         tuple of tuples: new values of (a, ya), (b, yb) and (c, yc)
         where c and yc are the newly discovered x and y values
    """
    a, ya = start
    b, yb = end
    s = last_yc

    # compute middle point using classical regula falsi update
    c = (yb * a - ya * b) / (yb - ya)
    yc = f(c)

    if yc * ya > 0:  # c is on the same side of f as the left limit (a)
        if yc * s > 0:
            m = 1 - yc / ya
            m = m if m > 0 else 0.5
            yb *= m
        a = c
        ya = yc
    else:  # c is on the same side of f as the right limit (b)
        if yc * s > 0:
            m = 1 - yc / yb
            m = m if m > 0 else 0.5
            ya *= m
        b = c
        yb = yc

    return (a, ya), (b, yb), (c, yc)


def _logL(beta, n, a, b, c0):
    """Compute the log-likelihood of the Poisson Background subtraction model

    Args:
        beta (float): background intensity
        n (numpy.ndarray): array of positive numbers
        a (numpy.ndarray): array of non-negative numbers (same length as n)
        b (numpy.ndarray): array of non-negative numbers (same length as n)
        c0 (float):

    Returns:
        float: log-likelihood at beta value

    """
    xi = a + b * beta
    xi[xi <= 0.0] = NUMERICAL_FLOOR
    logQ = gammainccln(n + 1, xi)
    logL = beta * c0 + np.sum(logQ)
    return logL


def _f_fprime(beta, n, a, b, c0):
    """Compute the derivative (f) and the second derivative (fprime) of logL

    Args:
        beta (float): background intensity
        n (numpy.ndarray): array of positive numbers
        a (numpy.ndarray): array of non-negative numbers (same length as n)
        b (numpy.ndarray): array of non-negative numbers (same length as n)
        c0 (float):

    Returns:
        tuple: (f, fprime) at beta value

    """
    xi = a + b * beta
    xi[xi <= 0.0] = NUMERICAL_FLOOR
    logQ = gammainccln(n + 1, xi)
    logH = n * np.log(xi) - xi - gammaln(n + 1) - logQ
    H = np.exp(logH)
    f = c0 - np.sum(b * H)
    fprime = - np.sum(b ** 2 * H * (n / xi + H - 1))

    return f, fprime


def _f_null(n, a, b, c0):
    """Compute f for beta = 0.0

    Args:
        n (numpy.ndarray): array of positive numbers
        a (numpy.ndarray): array of non-negative numbers (same length as n)
        b (numpy.ndarray): array of non-negative numbers (same length as n)
        c0 (float):

    Returns:
        float: f(0.0)
    """
    logQ = gammainccln(n + 1, a)
    if (a > 0.0).all():
        logH = n * np.log(a) - a - gammaln(n + 1) - logQ
        H = np.exp(logH)
    else:
        H = np.zeros_like(a)
        zeros = (a == 0)
        H[zeros] = 0.0  # assuming  n > 0
        H[~zeros] = np.exp(n[~zeros] * np.log(a[~zeros])
                           - a[~zeros]
                           - gammaln(n[~zeros] + 1)
                           - logQ[~zeros])

    f = c0 - np.sum(b * H)
    return f


def pbs_optimize(n, a, b, c0, beta_start=None, itermax=10, b_beta_tol=0.5):
    """Performs maximization of the PBS log-likelihood

    Log likelihood:
        logL(beta) = c0 * beta
                     + sum_{r: n_r > 0} logQ (n_r + 1,  a_r + b_r beta)

    If beta_start is None, the guess
    beta_guess = len(n) * np.median(n) / float(np.sum(b) - c0)
    is used as starting point.

    In any case, it is assumed that beta_start is between
    0.0 and beta_max, i.e. beta_max = np.sum(n) / float(np.sum(b) - c0)

    First Newton-Raphson algorithm is run. If this fails to get closer to
    the root of f = d/(d beta) logL, then Anderson-Bjoerck algorithm is run.

    Optimization terminates early, if b * beta changes less than b_beta_tol
    between iterations.

    Args:
        n: array of positives (zeros are not allowed)
        a: array of non-negatives (same length as n)
        b: array of positives (same length as n)
        c0: float
        beta_start: float, starting point of iteration
        itermax: maximum number of Anderson-Bjoerck iterations
        b_beta_tol: minimal value of the change of b * beta between iterations
            that is considered significant. If b * beta changes less than this
            between two iterations, the algorithm terminates early.

    Returns:
        float: new beta value
    """

    # define function the root of which we are looking for
    def _f(_beta):
        f_val, fprime_val = _f_fprime(_beta, n, a, b, c0)
        return f_val

    # If the f curve is decreasing even at 0.0, that is the optimum
    f_null = _f_null(n, a, b, c0)
    if f_null < 0.0:
        return 0.0

    betas = [0]
    fs = [f_null]

    # Compute upper limit (beta_max)
    beta_max = np.sum(n) / float(np.sum(b) - c0)
    for _ in range(10):
        f_at_beta_max = _f(beta_max)
        if f_at_beta_max < 0:
            break
        beta_max *= 2

    # Compute f and f' at starting point
    if beta_start is None:
        beta_start = len(n) * np.median(n) / float(np.sum(b) - c0)

    # Run Newton-Raphson algorithm
    beta = beta_start
    new_beta = beta
    for it in range(itermax):
        betas.append(beta)
        f, fprime = _f_fprime(beta, n, a, b, c0)
        fs.append(f)
        if fprime == 0.0:
            new_beta = -np.inf
            break

        new_beta = beta - f / fprime

        if (new_beta <= 0) or (new_beta >= beta_max):
            break
        if (np.abs(new_beta - beta) * b < b_beta_tol).all():
            return new_beta
        beta = new_beta
    f_start = fs[1]

    # Examine the results of N-R algorithm
    # If the new beta improves |f|, we accept it
    # otherwise, we use it as just another potential starting point
    # for Anderson-Bjoerck algorithm below
    if (new_beta > 0) and (new_beta < beta_max):
        f_new, fprime_new = _f_fprime(new_beta, n, a, b, c0)
        if np.abs(f_new) < np.abs(f_start):
            return new_beta
        betas.append(new_beta)
        fs.append(f_new)

    # Find the interval where f changes sign
    betas.append(beta_max)
    fs.append(f_at_beta_max)

    betas = np.array(betas)
    fs = np.array(fs)
    sort_order = np.argsort(betas)
    betas = betas[sort_order]
    fs = fs[sort_order]
    start = (betas[0], fs[0])
    end = (betas[len(fs) - 1], fs[len(fs) - 1])
    for idx_start, idx_end in zip(range(0, len(fs) - 1), range(1, len(fs))):
        if np.sign(fs[idx_start]) != np.sign(fs[idx_end]):
            start = (betas[idx_start], fs[idx_start])
            end = (betas[idx_end], fs[idx_end])
            break
    if start[1] * end[1] > 0:
        raise ValueError(f'f has the same sign at start and end, '
                         f'{start}, {end}')

    # Run Anderson-Bjoerck algorithm
    updated = (0.0, 0.0)
    for it in range(itermax):
        if start[0] == end[0]:
            return start[0]
        if start[1] == 0.0:
            return start[0]
        if end[1] == 0.0:
            return end[0]
        if start[1] * end[1] > 0:
            raise ValueError(
                f'f has the same sign at start and end, {start}, {end}')
        new_start, new_end, new_updated = anderson_bjoerk_step(_f, start, end,
                                                               updated[1])

        if (np.abs(new_updated[0] - updated[0]) * b < b_beta_tol).all():
            return new_updated[0]
        start = new_start
        end = new_end
        updated = new_updated

    return updated[0]
